use std::fs;
use std::path::PathBuf;
use std::process::{Command, Stdio};
use std::time::Duration;

use chrono::{DateTime, Local};
use serde::{Deserialize, Serialize};

use crate::api::{ChatChunk, ChatMessagePayload};

#[derive(Debug, Clone, Serialize, Deserialize)]
pub struct Session {
    session_id: String,
    ttl: DateTime<Local>,
    messages: Vec<ChatMessagePayload>,
    vqd: String
}

impl Session {
    pub fn terminal_session_id<T: From<String>>() -> T {
        #[cfg(unix)]
        fn inner() -> Result<String, Box<dyn std::error::Error>> {
            let child = Command::new("tty")
                .stdout(Stdio::piped())
                .stderr(Stdio::null())
                .spawn()?;

            let child = child.wait_with_output()?;
            if ! child.status.success() {
                Err("")? // value ignored
            }

            Ok(String::from_utf8(child.stdout)?)
        }

        #[cfg(not(unix))]
        fn inner() -> Result<String, ()> {
            Err(())
        }

        inner()
            .unwrap_or("default".into())
                .replace(['/', '\\', ':'], "-")
                .trim()
                .trim_start_matches('-')
                .to_string()
            .into()
    }

    fn path_for_id<T: Into<String>>(id: T) -> PathBuf {
        let id: String = id.into();
        std::env::temp_dir().join(format!("hey-duck-session-{id}"))
    }
    
    fn path(&self) -> PathBuf {
        Self::path_for_id(&self.session_id)
    }

    fn restore_with_id<T: Into<String>>(id: T) -> Option<Self> {
        fn inner(path: PathBuf) -> Result<Session, Box<dyn std::error::Error>> {
            if path.is_file() {
                let data = fs::read_to_string(path)?;
                Ok(serde_json::from_str(&data)?)
            } else {
                Err("")? // value ignored
            }
        }

        match inner(Self::path_for_id(id)) {
            Ok(mut session) => {
                if session.is_expired() {
                    session.destroy().expect("Couldn't destroy expired session");
                    None
                } else {
                    session.increase_ttl().expect("Couldn't increase TTL");
                    Some(session)
                }
            },
            Err(_) => None
        }
    }

    fn get_ttl() -> DateTime<Local> {
        Local::now() + Duration::from_secs(60 * 5)
    }

    pub fn increase_ttl(&mut self) -> Result<(), Box<dyn std::error::Error>> {
        self.ttl = Self::get_ttl();
        Ok(self.save()?)
    }

    fn new<I: Into<String>, V: Into<String>>(id: I, vqd: V) -> Self {
        Self {
            session_id: id.into(),
            ttl: Self::get_ttl(),
            messages: vec![],
            vqd: vqd.into()
        }
    }

    fn save(&self) -> Result<(), Box<dyn std::error::Error>> {
        fs::write(self.path(), serde_json::to_string_pretty(self)?)?;
        Ok(())
    }

    fn is_expired(&self) -> bool {
        self.ttl < Local::now()
    }

    pub fn destroy(&self) -> Result<(), Box<dyn std::error::Error>> {
        Ok(fs::remove_file(self.path())?)
    }

    pub fn create_or_restore<T: Into<String>>(vqd: T) -> Self {
        let session_id: String = Self::terminal_session_id();
        match Self::restore_with_id(&session_id) {
            Some(session) => session,
            None => {
                let session = Self::new(&session_id, vqd);
                session.save().expect("Couldn't save new session");
                session
            }
        }
    }

    pub fn push_message(&mut self, msg: &ChatMessagePayload) -> Result<(), Box<dyn std::error::Error>> {
        self.messages.push(msg.clone());
        Ok(self.save()?)
    }

    pub fn push_ai_message_chunk(&mut self, chunk: &ChatChunk) -> Result<(), Box<dyn std::error::Error>> {
        if self.messages.last().unwrap().is_ai() {
            self.messages.last_mut().unwrap().content += chunk.message.as_str();
        } else {
            self.messages.push(chunk.clone().into());
        }

        Ok(self.save()?)
    }

    pub fn get_messages(&self) -> Vec<ChatMessagePayload> {
        self.messages.clone()
    }

    pub fn restore_vqd() -> Option<String> {
       Self::restore_with_id(Self::terminal_session_id::<String>()).map(|x| x.vqd)
    }

    pub fn set_last_vqd<T: Into<String>>(&mut self, vqd: T) -> Result<(), Box<dyn std::error::Error>> {
        self.vqd = vqd.into();
        Ok(self.save()?)
    }
}